-- ============================================================= --
-- FOG CONTROL MOD
-- Autor: BMprod
-- date: 20.05.2025
-- dateUpdate: 02.06.2025
-- https://vk.com/biostilmodprod
-- ============================================================= --
FogControl = {}
FogControl.name = g_currentModName
FogControl.path = g_currentModDirectory

FogControl.SETTINGS = {}
FogControl.CONTROLS = {}

addModEventListener(FogControl)

FogControl.menuItems = {
    'fogEnabled',
    'groundFogDensity',
    'heightFogDensity',
    'fogCoverage',
    'fogStartTime',
    'fogEndTime',
    'fogWeatherTypes',
    'fogSeasons',
}

-- Default settings
FogControl.fogEnabled = true
FogControl.groundFogDensity = 0.01
FogControl.heightFogDensity = 0.1
FogControl.fogCoverage = 0.05
FogControl.fogStartTime = 240
FogControl.fogEndTime = 600
FogControl.fogWeatherTypes = {
    [WeatherType.CLOUDY] = true,
    [WeatherType.RAIN] = true,
}
FogControl.fogSeasons = {
    [Season.SPRING] = true,
    [Season.SUMMER] = true,
    [Season.AUTUMN] = true,
}
FogControl.lastCheckedDay = -1
FogControl.settingsAppliedForDay = false

-- Helper function to get table keys
function table.keys(tbl)
    local keys = {}
    for key, _ in pairs(tbl or {}) do
        table.insert(keys, tostring(key))
    end
    return keys
end

function FogControl:updateFogSettings()
    if not g_currentMission or not g_currentMission.environment or not g_currentMission.environment.weather then
        return
    end

    local fogUpdater = g_currentMission.environment.weather.fogUpdater
    local fogSettings = FogSettings.new()

    if not FogControl.fogEnabled then
        fogSettings.groundFogGroundLevelDensity = 0
        fogSettings.heightFogGroundLevelDensity = 0
        fogSettings.groundFogCoverageEdge0 = 0
        fogSettings.groundFogCoverageEdge1 = 0
        fogSettings.groundFogExtraHeight = 0
        fogSettings.groundFogMinValleyDepth = 0
        fogSettings.heightFogMaxHeight = 0
        fogSettings.groundFogStartDayTimeMinutes = 0
        fogSettings.groundFogEndDayTimeMinutes = 0
        fogSettings.groundFogWeatherTypes = {}
        fogUpdater.visibilityAlpha = 0
    else
        fogSettings.groundFogGroundLevelDensity = FogControl.groundFogDensity
        fogSettings.heightFogGroundLevelDensity = FogControl.heightFogDensity
        fogSettings.groundFogCoverageEdge0 = FogControl.fogCoverage
        fogSettings.groundFogCoverageEdge1 = math.max(FogControl.fogCoverage + 0.1, 1.0)
        fogSettings.groundFogExtraHeight = 7
        fogSettings.groundFogMinValleyDepth = 1.5
        fogSettings.heightFogMaxHeight = 550
        fogSettings.groundFogStartDayTimeMinutes = FogControl.fogStartTime
        fogSettings.groundFogEndDayTimeMinutes = FogControl.fogEndTime
        fogSettings.groundFogWeatherTypes = table.clone(FogControl.fogWeatherTypes or {})
        fogUpdater.visibilityAlpha = 1
    end

    fogUpdater:setTargetFog(fogSettings, 30000)
    fogUpdater.isDirty = true

    if g_server ~= nil then
        g_server:broadcastEvent(FogStateEvent.new(
            fogSettings.groundFogGroundLevelDensity,
            fogUpdater.lastMieScale,
            fogUpdater.alpha,
            30000,
            g_currentMission.environment.weather.nightFactor,
            g_currentMission.environment.weather.dayFactor
        ))
    end
end

function FogControl:update(dt)
    if not g_currentMission or not g_currentMission:getIsServer() then
        return
    end

    if not FogControl.initialised then
        FogControl:updateFogSettings()
        FogControl.initialised = true
    end

    local environment = g_currentMission.environment
    if environment and environment.weather then
        local currentDay = environment.weather.owner.currentMonotonicDay
        local currentDayTime = environment.dayTime / 60000
        local currentSeason = environment.currentSeason

        if currentDay ~= FogControl.lastCheckedDay then
            FogControl.settingsAppliedForDay = false
            FogControl.lastCheckedDay = currentDay
        end

        if not FogControl.settingsAppliedForDay and FogControl.fogEnabled and FogControl.fogSeasons[currentSeason] then
            local fogStartTime = FogControl.fogStartTime
            local timeUntilStart = fogStartTime - currentDayTime
            if timeUntilStart > 0 and timeUntilStart <= 1 then
                FogControl:updateFogSettings()
                FogControl.settingsAppliedForDay = true
            end
        end
    end
end

function FogControl:loadMap(name)
    FogControl:readSettings()
    addConsoleCommand("gsFogControl", "Control fog settings", "consoleCommandFogControl", FogControl)
    FogControl:updateFogSettings()
end

function FogControl:onDayChanged()
    FogControl.settingsAppliedForDay = false
    FogControl.lastCheckedDay = g_currentMission.environment.weather.owner.currentMonotonicDay
end

function FogControl:consoleCommandFogControl(param1, param2)
    if param1 == "enable" then
        local value = param2:lower() == "true"
        FogControl.fogEnabled = value
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog enabled: %s", tostring(value))
    elseif param1 == "groundDensity" then
        local value = tonumber(param2) or FogControl.groundFogDensity
        FogControl.groundFogDensity = math.clamp(value, 0.0, 1.0)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Ground fog density: %.4f", FogControl.groundFogDensity)
    elseif param1 == "heightDensity" then
        local value = tonumber(param2) or FogControl.heightFogDensity
        FogControl.heightFogDensity = math.clamp(value, 0.0, 1.0)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Height fog density: %.4f", FogControl.heightFogDensity)
    elseif param1 == "coverage" then
        local value = tonumber(param2) or FogControl.fogCoverage
        FogControl.fogCoverage = math.clamp(value, 0.0, 1.0)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog coverage: %.4f", FogControl.fogCoverage)
    elseif param1 == "startTime" then
        local value = tonumber(param2) or FogControl.fogStartTime
        FogControl.fogStartTime = math.clamp(value, 0, 1440)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog start time: %d minutes", FogControl.fogStartTime)
    elseif param1 == "endTime" then
        local value = tonumber(param2) or FogControl.fogEndTime
        FogControl.fogEndTime = math.clamp(value, 0, 1440)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog end time: %d minutes", FogControl.fogEndTime)
    elseif param1 == "weatherTypes" then
        local types = string.split(param2, ",")
        FogControl.fogWeatherTypes = {}
        for _, typeName in ipairs(types) do
            local weatherType = WeatherType.getByName(typeName:upper())
            if weatherType then
                FogControl.fogWeatherTypes[weatherType] = true
            end
        end
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog weather types: %s", table.concat(table.keys(FogControl.fogWeatherTypes), ", "))
    elseif param1 == "seasons" then
        local seasons = string.split(param2, ",")
        FogControl.fogSeasons = {}
        for _, seasonName in ipairs(seasons) do
            local season = Season.getByName(seasonName:upper())
            if season then
                FogControl.fogSeasons[season] = true
            end
        end
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog seasons: %s", table.concat(table.keys(FogControl.fogSeasons), ", "))
    end
    return "Usage: gsFogControl [enable|groundDensity|heightDensity|coverage|startTime|endTime|weatherTypes|seasons] <value>"
end

function FogControl:setValue(id, value)
    FogControl[id] = value
end

function FogControl:getValue(id)
    return FogControl[id]
end

function FogControl:getStateIndex(id, value)
    local value = value or FogControl:getValue(id)
    local setting = FogControl.SETTINGS[id]
    if not setting then
        return 1
    end
    local values = setting.values
    if type(value) == 'number' then
        local index = setting.default
        local initialDiff = math.huge
        for i, v in pairs(values) do
            local currentDiff = math.abs(v - value)
            if currentDiff < initialDiff then
                initialDiff = currentDiff
                index = i
            end
        end
        return index
    elseif type(value) == 'table' then
        local count = 0
        for _, v in pairs(value) do
            if v then count = count + 1 end
        end
        return math.min(count + 1, #values)
    else
        for i, v in pairs(values) do
            if v == value then
                return i
            end
        end
    end
    return setting.default
end

function FogControl:writeSettings()
    local key = "fogControlSettings"
    local userSettingsFile = Utils.getFilename("modSettings/FogControl.xml", getUserProfileAppPath())
    local xmlFile = createXMLFile("settings", userSettingsFile, key)

    if xmlFile ~= 0 then
        for _, id in pairs(FogControl.menuItems) do
            local value = FogControl:getValue(id)
            local xmlValueKey = "fogControlSettings." .. id .. "#value"
            if type(value) == 'number' then
                setXMLFloat(xmlFile, xmlValueKey, value)
            elseif type(value) == 'boolean' then
                setXMLBool(xmlFile, xmlValueKey, value)
            elseif type(value) == 'table' then
                local strValue = table.concat(table.keys(value), ",")
                setXMLString(xmlFile, xmlValueKey, strValue)
            end
        end
        saveXMLFile(xmlFile)
        delete(xmlFile)
    end
end

function FogControl:readSettings()
    local userSettingsFile = Utils.getFilename("modSettings/FogControl.xml", getUserProfileAppPath())
    if not fileExists(userSettingsFile) then
        FogControl:writeSettings()
        return
    end

    local xmlFile = loadXMLFile("fogControlSettings", userSettingsFile)
    if xmlFile ~= 0 then
        for _, id in pairs(FogControl.menuItems) do
            local setting = FogControl.SETTINGS[id]
            if setting then
                local xmlValueKey = "fogControlSettings." .. id .. "#value"
                if hasXMLProperty(xmlFile, xmlValueKey) then
                    local value = FogControl:getValue(id)
                    if type(value) == 'number' then
                        value = getXMLFloat(xmlFile, xmlValueKey) or value
                    elseif type(value) == 'boolean' then
                        value = getXMLBool(xmlFile, xmlValueKey) or value
                    elseif type(value) == 'table' then
                        local strValue = getXMLString(xmlFile, xmlValueKey)
                        value = {}
                        if strValue then
                            local items = string.split(strValue, ",")
                            for _, item in ipairs(items) do
                                if id == "fogWeatherTypes" then
                                    local weatherType = WeatherType.getByName(item:upper())
                                    if weatherType then value[weatherType] = true end
                                elseif id == "fogSeasons" then
                                    local season = Season.getByName(item:upper())
                                    if season then value[season] = true end
                                end
                            end
                        end
                    end
                    FogControl:setValue(id, value)
                end
            end
        end
        delete(xmlFile)
    end
end

-- Menu integration
FogControl.SETTINGS.fogEnabled = {
    default = 1,
    values = {false, true},
    strings = {g_i18n:getText("ui_off"), g_i18n:getText("ui_on")}
}

local percentValues = {}
local percentStrings = {}
for i = 0, 1000 do
    percentValues[i + 1] = i / 1000
    percentStrings[i + 1] = string.format("%.1f%%", i / 10)
end

FogControl.SETTINGS.groundFogDensity = {
    default = 15,
    values = percentValues,
    strings = percentStrings
}

FogControl.SETTINGS.heightFogDensity = {
    default = 200,
    values = percentValues,
    strings = percentStrings
}

FogControl.SETTINGS.fogCoverage = {
    default = 100,
    values = percentValues,
    strings = percentStrings
}

FogControl.SETTINGS.fogStartTime = {
    default = 2,
    values = {0, 240, 300, 360, 420},
    strings = {"00:00", "04:00", "05:00", "06:00", "07:00"}
}

FogControl.SETTINGS.fogEndTime = {
    default = 3,
    values = {480, 540, 600, 660, 720},
    strings = {"08:00", "09:00", "10:00", "11:00", "12:00"}
}

FogControl.SETTINGS.fogWeatherTypes = {
    default = 1,
    values = {
        {},
        {[WeatherType.CLOUDY] = true},
        {[WeatherType.CLOUDY] = true, [WeatherType.RAIN] = true},
        {[WeatherType.CLOUDY] = true, [WeatherType.RAIN] = true, [WeatherType.SNOW] = true},
    },
    strings = {
        g_i18n:getText("ui_none"),
        "CLOUDY",
        "CLOUDY, RAIN",
        "CLOUDY, RAIN, SNOW"
    }
}

FogControl.SETTINGS.fogSeasons = {
    default = 3,
    values = {
        {},
        {[Season.SPRING] = true},
        {[Season.SPRING] = true, [Season.AUTUMN] = true},
        {[Season.SPRING] = true, [Season.SUMMER] = true, [Season.AUTUMN] = true},
        {[Season.SPRING] = true, [Season.AUTUMN] = true, [Season.WINTER] = true},
        {[Season.SPRING] = true, [Season.SUMMER] = true, [Season.AUTUMN] = true, [Season.WINTER] = true},
    },
    strings = {
        g_i18n:getText("ui_none"),
        "SPRING",
        "SPRING, AUTUMN",
        "SPRING, SUMMER, AUTUMN",
        "SPRING, AUTUMN, WINTER",
        "SPRING, SUMMER, AUTUMN, WINTER"
    }
}

function FogControl:addMenuOption(id)
    local inGameMenu = g_gui.screenControllers[InGameMenu]
    local settingsPage = inGameMenu.pageSettings
    local settingsLayout = settingsPage.generalSettingsLayout

    local original = #FogControl.SETTINGS[id].values == 2 and settingsPage.checkWoodHarvesterAutoCutBox or settingsPage.multiVolumeVoiceBox
    local options = FogControl.SETTINGS[id].strings

    local menuOptionBox = original:clone(settingsLayout)
    menuOptionBox.id = id .. "box"

    local menuOption = menuOptionBox.elements[1]
    menuOption.id = id
    menuOption.target = FogControl
    menuOption:setCallback("onClickCallback", "onMenuOptionChanged")
    menuOption:setDisabled(false)

    local toolTip = menuOption.elements[1]
    toolTip:setText(g_i18n:getText("tooltip_fogcontrol_" .. id))

    local setting = menuOptionBox.elements[2]
    setting:setText(g_i18n:getText("setting_fogcontrol_" .. id))

    menuOption:setTexts({unpack(options)})
    menuOption:setState(FogControl:getStateIndex(id))

    FogControl.CONTROLS[id] = menuOption

    local function updateFocusIds(element)
        if not element then return end
        element.focusId = FocusManager:serveAutoFocusId()
        for _, child in pairs(element.elements) do
            updateFocusIds(child)
        end
    end
    updateFocusIds(menuOptionBox)
    table.insert(settingsPage.controlsList, menuOptionBox)
end

function FogControl:onMenuOptionChanged(state, menuOption)
    local id = menuOption.id
    local value = FogControl.SETTINGS[id].values[state]
    FogControl:setValue(id, value)
    FogControl:writeSettings()
    FogControl:updateFogSettings()
end

-- Add menu section and options
local inGameMenu = g_gui.screenControllers[InGameMenu]
local settingsPage = inGameMenu.pageSettings
local settingsLayout = settingsPage.generalSettingsLayout

local sectionTitle
for _, elem in ipairs(settingsLayout.elements) do
    if elem.name == "sectionHeader" then
        sectionTitle = elem:clone(settingsLayout)
        break
    end
end
if sectionTitle then
    sectionTitle:setText(g_i18n:getText("menu_FogControl_TITLE"))
else
    local title = TextElement.new()
    title:applyProfile("fs25_settingsSectionHeader", true)
    title:setText(g_i18n:getText("menu_FogControl_TITLE"))
    title.name = "sectionHeader"
    settingsLayout:addElement(title)
    sectionTitle = title
end

sectionTitle.focusId = FocusManager:serveAutoFocusId()
table.insert(settingsPage.controlsList, sectionTitle)
FogControl.CONTROLS[sectionTitle.name] = sectionTitle

for _, id in pairs(FogControl.menuItems) do
    FogControl:addMenuOption(id)
end
settingsLayout:invalidateLayout()

-- Handle focus for custom controls
FocusManager.setGui = Utils.appendedFunction(FocusManager.setGui, function(_, gui)
    if gui == "ingameMenuSettings" then
        for _, control in pairs(FogControl.CONTROLS) do
            if not control.focusId or not FocusManager.currentFocusData.idToElementMapping[control.focusId] then
                if not FocusManager:loadElementFromCustomValues(control, nil, nil, false, false) then
                    Logging.warning("Could not register control %s with the focus manager", control.id or control.name or control.focusId)
                end
            end
        end
        local settingsPage = g_gui.screenControllers[InGameMenu].pageSettings
        settingsPage.generalSettingsLayout:invalidateLayout()
    end
end)

InGameMenuSettingsFrame.onFrameOpen = Utils.appendedFunction(InGameMenuSettingsFrame.onFrameOpen, function()
    local isAdmin = g_currentMission:getIsServer() or g_currentMission.isMasterUser
    for _, id in pairs(FogControl.menuItems) do
        local menuOption = FogControl.CONTROLS[id]
        menuOption:setState(FogControl:getStateIndex(id))
        menuOption:setDisabled(not isAdmin)
    end
end)

-- Register for day change event
Environment.onDayChanged = Utils.appendedFunction(Environment.onDayChanged, FogControl.onDayChanged)

-- Override Weather.randomizeFog to apply FogControl settings
Weather.randomizeFog = Utils.overwrittenFunction(Weather.randomizeFog, function(self, superFunc, duration)
    FogControl:updateFogSettings()
end)

-- Override FogUpdater.update to fully control fog behavior
FogUpdater.update = Utils.overwrittenFunction(FogUpdater.update, function(self, superFunc, dt)
    if not FogControl.fogEnabled then
        self.currentFog.groundFogGroundLevelDensity = 0
        self.currentFog.heightFogGroundLevelDensity = 0
        self.currentFog.groundFogCoverageEdge0 = 0
        self.currentFog.groundFogCoverageEdge1 = 0
        self.visibilityAlpha = 0
        self.isDirty = true
    else
        if not FogControl.fogSeasons[g_currentMission.environment.currentSeason] then
            self.visibilityAlpha = 0
            self.isDirty = true
            return
        end

        if self.alpha ~= 1 then
            local newAlpha = self.alpha + dt / self.duration
            self.alpha = math.min(newAlpha, 1)
            local alpha = self.alpha
            local currentFog = self.currentFog
            local lastFog = self.lastFog
            local targetFog = self.targetFog
            currentFog.groundFogCoverageEdge0 = MathUtil.lerp(lastFog.groundFogCoverageEdge0, targetFog.groundFogCoverageEdge0, alpha)
            currentFog.groundFogCoverageEdge1 = MathUtil.lerp(lastFog.groundFogCoverageEdge1, targetFog.groundFogCoverageEdge1, alpha)
            currentFog.groundFogExtraHeight = MathUtil.lerp(lastFog.groundFogExtraHeight, targetFog.groundFogExtraHeight, alpha)
            currentFog.groundFogGroundLevelDensity = MathUtil.lerp(lastFog.groundFogGroundLevelDensity, targetFog.groundFogGroundLevelDensity, alpha)
            currentFog.groundFogMinValleyDepth = MathUtil.lerp(lastFog.groundFogMinValleyDepth, targetFog.groundFogMinValleyDepth, alpha)
            currentFog.heightFogMaxHeight = MathUtil.lerp(lastFog.heightFogMaxHeight, targetFog.heightFogMaxHeight, alpha)
            currentFog.heightFogGroundLevelDensity = MathUtil.lerp(lastFog.heightFogGroundLevelDensity, targetFog.heightFogGroundLevelDensity, alpha)
            self.isDirty = true
        end

        local isFogPossible = self:getIsFogPossible()
        self.visibilityAlpha = isFogPossible and 1 or 0
    end

    if self.isDirty then
        local currentFog = self.currentFog
        local coverageEdge0 = MathUtil.lerp(1, currentFog.groundFogCoverageEdge0, self.visibilityAlpha)
        local coverageEdge1 = MathUtil.lerp(1, currentFog.groundFogCoverageEdge1, self.visibilityAlpha)
        setGroundFogGlobalCoverage(coverageEdge0, coverageEdge1)
        setGroundFogHeight(currentFog.groundFogExtraHeight)
        setGroundFogGroundLevelDensity(currentFog.groundFogGroundLevelDensity)
        setGroundFogMinimumValleyDepth(currentFog.groundFogMinValleyDepth)
        setHeightFogGroundLevelDensity(currentFog.heightFogGroundLevelDensity)
        setHeightFogMaxHeight(currentFog.heightFogMaxHeight)
        self.isDirty = false
    end
end)

-- Override FogUpdater.getIsFogPossible to use FogControl settings
FogUpdater.getIsFogPossible = Utils.overwrittenFunction(FogUpdater.getIsFogPossible, function(self, superFunc)
    if not FogControl.fogEnabled or not FogControl.fogSeasons[g_currentMission.environment.currentSeason] then
        return false
    end

    local environment = g_currentMission.environment
    if environment == nil then
        return false
    end

    local currentWeatherType = environment.weather:getCurrentWeatherType()
    if not FogControl.fogWeatherTypes[currentWeatherType] then
        return false
    end

    local currentDayTimeMinutes = environment.dayTime / 1000 / 60
    return not MathUtil.getIsOutOfBounds(currentDayTimeMinutes, FogControl.fogStartTime, FogControl.fogEndTime)
end)

-- Override FogUpdater.setTargetFog to enforce FogControl settings
FogUpdater.setTargetFog = Utils.overwrittenFunction(FogUpdater.setTargetFog, function(self, superFunc, fogSettings, duration)
    if not FogControl.fogEnabled then
        local disabledFog = FogSettings.new()
        disabledFog.groundFogGroundLevelDensity = 0
        disabledFog.heightFogGroundLevelDensity = 0
        disabledFog.groundFogCoverageEdge0 = 0
        disabledFog.groundFogCoverageEdge1 = 0
        disabledFog.groundFogExtraHeight = 0
        disabledFog.groundFogMinValleyDepth = 0
        disabledFog.heightFogMaxHeight = 0
        disabledFog.groundFogStartDayTimeMinutes = 0
        disabledFog.groundFogEndDayTimeMinutes = 0
        disabledFog.groundFogWeatherTypes = {}
        self.alpha = 0
        self.duration = math.max(1, duration or 30000)
        self.lastFog = self.targetFog
        self.targetFog = disabledFog
        self.currentFog.groundFogWeatherTypes = {}
        return
    end

    local customFog = FogSettings.new()
    customFog.groundFogGroundLevelDensity = FogControl.groundFogDensity
    customFog.heightFogGroundLevelDensity = FogControl.heightFogDensity
    customFog.groundFogCoverageEdge0 = FogControl.fogCoverage
    customFog.groundFogCoverageEdge1 = math.max(FogControl.fogCoverage + 0.1, 1.0)
    customFog.groundFogExtraHeight = 7
    customFog.groundFogMinValleyDepth = 1.5
    customFog.heightFogMaxHeight = 550
    customFog.groundFogStartDayTimeMinutes = FogControl.fogStartTime
    customFog.groundFogEndDayTimeMinutes = FogControl.fogEndTime
    customFog.groundFogWeatherTypes = table.clone(FogControl.fogWeatherTypes or {})
    self.alpha = 0
    self.duration = math.max(1, duration or 30000)
    self.lastFog = self.targetFog
    self.targetFog = customFog
    self.currentFog.groundFogWeatherTypes = table.clone(FogControl.fogWeatherTypes or {})
end)